export interface SineBellOptions {
  /**
   * Specifies the exponent of the sine-bell; Non-integer values are allowed.
   * Common values are 1.0 (for ordinary sine-bell) and 2.0 (for squared-bell functions).
   * @default 1
   */
  exponent?: number;
  /**
   * Specifies the starting point of the sine-bell in units of pi radians.
   * Common values are 0.0 (for a sine window which starts height at 0.0) and
   * 0.5 (for a cosine window, which starts at height 1.0).
   * @default 0
   */
  offset?: number;
  /**
   * Specifies the ending point of the sine-bell in units of pi radians.
   * Common values are 1.0 (for a window which goes to 0.0 height at the last point) and
   * 0.95 (for a window which doesn't go all the way to 0.0).
   * @default 1
   */
  end?: number;
  /**
   * Specifies the number of points in the window function.
   */
  length: number;
}

export function sineBell(options: SineBellOptions) {
  const { exponent = 1, offset = 0, end = 1, length } = options;

  if (length === undefined) {
    throw new Error('length options should exists for sineBell shape');
  }

  if (end < 0 || end > 1) {
    throw new Error(`the end parameter should be [0-1]`);
  }
  const c1 = Math.PI * offset;
  const c2 = (Math.PI * (end - offset)) / (length - 1);
  return (i: number) => Math.pow(Math.sin(c1 + c2 * i), exponent);
}
