import { toBeDeepCloseTo, toMatchCloseTo } from 'jest-matcher-deep-close-to';
import OCL from 'openchemlib/full';

import { getAssignments, SpectraData } from '../getAssignments';
import {
  NMRSignal1DWithId,
  NMRSignal2DWithId,
  SpectraData1DWithIds,
  SpectraData2DWithIds,
  SpectraDataWithIds,
} from '../nmrAssigment';
import { isSpectraData1D } from '../utils/generalAssignment/isSpectraData1D';
import { searchIndices } from '../utils/generalAssignment/searchIndices';

import ethylbenzeneData from './data/ethylBenzeneData.json';
import resultEthylbenzene from './data/ethylBenzeneResult1D.json';
import ethylbenzeneCarbonPrediction from './data/ethylbenzenePrediction_C.json';
import ethylbenzeneProtonPrediction from './data/ethylbenzenePrediction_H.json';

function isSignal1D(
  signal: NMRSignal2DWithId | NMRSignal1DWithId,
): signal is NMRSignal1DWithId {
  return 'delta' in signal;
}

expect.extend({ toBeDeepCloseTo, toMatchCloseTo });

function checkMatchIDs(
  assignment: SpectraDataWithIds[],
  assignmentResult: SpectraDataWithIds[],
): void {
  for (const spectrum of assignment) {
    const data = isSpectraData1D(spectrum) ? spectrum.ranges : spectrum.zones;
    for (const dataRZ of data) {
      for (const signal of dataRZ.signals) {
        const signalId = signal.id;
        const { spectrumIndex, elementIndex, signalIndex } = searchIndices(
          signalId,
          assignmentResult,
        );
        if (isSignal1D(signal)) {
          if (!signal.diaIDs) continue;
          const spectrumResult = assignmentResult[
            spectrumIndex
          ] as SpectraData1DWithIds;
          const signalResult =
            spectrumResult.ranges[elementIndex].signals[signalIndex];
          expect(signal.diaIDs).toMatchCloseTo(signalResult.diaIDs);
        } else {
          const axis = ['x', 'y'] as Array<'x' | 'y'>;
          for (const axe of axis) {
            if (!signal[axe].diaIDs) continue;
            const spectrumResult = assignmentResult[
              spectrumIndex
            ] as SpectraData2DWithIds;
            const signalResult =
              spectrumResult.zones[elementIndex].signals[signalIndex];
            expect(signal[axe].diaIDs).toMatchCloseTo(signalResult[axe].diaIDs);
          }
        }
      }
    }
  }
}

describe('automatic assignment', () => {
  const molfile = String(ethylbenzeneData.molfile);
  const spectra = ethylbenzeneData.spectra as SpectraData[];
  const molecule = OCL.Molecule.fromMolfile(molfile);
  it('simple assignment', async () => {
    const result = await getAssignments(
      { molecule, spectra },
      {
        justAssign: [['H'], ['C']],
        nbAllowedUnAssigned: { H: 0, C: 1 },
        minScore: 0.01,
        maxSolutions: 30,
        predictions: {
          C: ethylbenzeneCarbonPrediction,
          H: ethylbenzeneProtonPrediction,
        },
        restrictionByCS: {
          tolerance: { H: 0.2, C: 1 },
          useChemicalShiftScore: true,
          chemicalShiftRestriction: true,
        },
      },
    );
    const result2 = await getAssignments(
      { molecule, spectra },
      {
        justAssign: [['C'], ['H']],
        nbAllowedUnAssigned: { H: 0, C: 1 },
        minScore: 0.01,
        maxSolutions: 1000,
        predictions: {
          C: ethylbenzeneCarbonPrediction,
          H: ethylbenzeneProtonPrediction,
        },
        restrictionByCS: {
          tolerance: { H: 0.2, C: 1 },
          useChemicalShiftScore: true,
          chemicalShiftRestriction: true,
        },
      },
    );
    const result3 = await getAssignments(
      { molecule, spectra },
      {
        nbAllowedUnAssigned: { H: 0, C: 1 },
        minScore: 0.01,
        maxSolutions: 100,
        predictions: {
          C: ethylbenzeneCarbonPrediction,
          H: ethylbenzeneProtonPrediction,
        },
        restrictionByCS: {
          tolerance: { H: 0.2, C: 1 },
          useChemicalShiftScore: true,
          chemicalShiftRestriction: true,
        },
      },
    );

    const result4 = await getAssignments(
      { molecule, spectra },
      {
        justAssign: [['H', 'C']],
        nbAllowedUnAssigned: { H: 0, C: 1 },
        minScore: 0.01,
        maxSolutions: 1,
        predictions: {
          C: ethylbenzeneCarbonPrediction,
          H: ethylbenzeneProtonPrediction,
        },
        restrictionByCS: {
          tolerance: { H: 0.2, C: 1 },
          useChemicalShiftScore: true,
          chemicalShiftRestriction: true,
        },
      },
    );

    checkMatchIDs(
      resultEthylbenzene.spectra as SpectraDataWithIds[],
      result[0].assignment,
    );
    checkMatchIDs(
      resultEthylbenzene.spectra as SpectraDataWithIds[],
      result2[0].assignment,
    );
    checkMatchIDs(
      resultEthylbenzene.spectra as SpectraDataWithIds[],
      result3[0].assignment,
    );
    checkMatchIDs(
      resultEthylbenzene.spectra as SpectraDataWithIds[],
      result4[0].assignment,
    );
  });
});
