import { v4 } from '@lukeed/uuid';
import { NmrData1D, NmrData2D } from 'cheminfo-types';

import { BaseFilter, Filter } from '../types/Filter';
import { cloneData1D } from '../utilities/cloneData1D';
import { cloneData2D } from '../utilities/cloneData2D';

import { Filters } from './Filters';
import { Entry1D } from './data1d/Entry1D';
import { updateIntegrals } from './data1d/utils/updateIntegrals';
import { updatePeaks } from './data1d/utils/updatePeaks';
import { updateRanges } from './data1d/utils/updateRanges';
import { Entry2D } from './data2d/Entry2D';
import { updateZones } from './data2d/utils/updateZones';

export interface FilterDomainUpdateRules {
  updateYDomain: boolean;
  updateXDomain: boolean;
}

export type Entry = Entry1D | Entry2D;
function resetDataToOrigin(datum: Entry1D | Entry2D) {
  datum.data =
    datum.info.dimension === 1
      ? cloneData1D(datum.originalData as NmrData1D)
      : cloneData2D(datum.originalData as NmrData2D);

  const { originalInfo } = datum;
  if (originalInfo) {
    datum.info = { ...originalInfo };
  }
}
// update the the part of the data that affected by applying the filters
function updateData(datum: Entry1D | Entry2D) {
  if (datum.info.dimension === 1) {
    updatePeaks(datum as Entry1D);
    updateRanges(datum as Entry1D);
    updateIntegrals(datum as Entry1D);
  } else if (datum.info.dimension === 2) {
    updateZones(datum as Entry2D);
  }
}

interface ApplyFilterOptions {
  forceReapply?: boolean;
  filterIndex?: number;
}

interface FilterOption {
  name: string;
  label: string;
  value: any;
}

/***
 * @param {object} Filters [{name:'',options:{}},{...}]
 */
export function applyFilter(
  datum: Entry1D | Entry2D,
  filters: BaseFilter[] = [],
  options: ApplyFilterOptions = {},
) {
  const { forceReapply = false, filterIndex } = options;
  let isReduced = false;
  for (const filter of filters) {
    const filterOption = {
      name: filter.name,
      label: Filters[filter.name].name,
      value: filter.value,
    };
    const previousFilter = lookupForFilter(datum, filter.name);
    const isDeleteAllow =
      'isDeleteAllow' in filter ? filter.isDeleteAllow : true;
    if (previousFilter) {
      const reduceResult = Filters[filter.name].reduce(
        previousFilter.value,
        filterOption.value,
      );
      if (reduceResult.once) {
        if (!isReduced) {
          isReduced = true;
        }
        if (reduceResult.reduce != null && reduceResult.reduce !== undefined) {
          replaceFilter(datum, previousFilter.id, reduceResult.reduce);
        }
      } else {
        addFilter(datum, filterOption, { isDeleteAllow, filterIndex });
      }
    } else {
      addFilter(datum, filterOption, { isDeleteAllow, filterIndex });
    }
  }
  if (forceReapply) {
    reapplyFilters(datum);
  } else if (isReduced) {
    if (filters.length === 1 && isLastFilter(datum, filters[0].name)) {
      Filters[filters[0].name].apply(datum, filters[0].value);
    } else {
      reapplyFilters(datum);
    }
  } else {
    for (const filter of filters) {
      Filters[filter.name].apply(datum, filter.value);
    }
  }
  updateData(datum);
}

function isLastFilter(datum: Entry1D | Entry2D, id: string) {
  const index = datum.filters.findIndex((f) => f.name === id);
  if (datum.filters.length === index + 1) {
    return true;
  }
  return false;
}

export function lookupForFilter(datum: Entry1D | Entry2D, filterName: string) {
  return datum.filters.find((f) => f.name === filterName);
}

export function reapplyFilters(datum: Entry1D | Entry2D, filters?: Filter[]) {
  const _filters = filters || datum.filters;
  enableFilter(datum, null, null, _filters);
}

// id filter id
export function enableFilter(
  datum: Entry1D | Entry2D,
  id: string | null,
  checked: boolean | null,
  filters?: Filter[],
) {
  datum.filters = datum.filters.slice(0);
  if (id) {
    datum.filters = datum.filters.map((filter) => {
      return { ...filter, flag: filter.id === id ? checked : filter.flag };
    });
  }

  resetDataToOrigin(datum);

  const _filters = filters || datum.filters;

  for (let filterIndex = 0; filterIndex < _filters.length; filterIndex++) {
    const filter = datum.filters[filterIndex];
    datum.filters[filterIndex] = {
      ...datum.filters[filterIndex],
    };

    if (filter.flag) {
      try {
        Filters[filter.name].apply(datum, filter.value);
      } catch (error: any) {
        datum.filters[filterIndex] = {
          ...datum.filters[filterIndex],
          error: error.message,
        };
      }
    }
  }
}

export function deleteFilter(datum: Entry1D | Entry2D, id?: string) {
  // delete the specified filter otherwise delete all filters where isDeleteAllow is true
  datum.filters = datum.filters
    .slice(0)
    .filter((filter) => (id ? filter.id !== id : !filter.isDeleteAllow));

  resetDataToOrigin(datum);

  for (let i = 0; i < datum.filters.length; i++) {
    datum.filters[i] = {
      ...datum.filters[i],
    };

    if (datum.filters[i].flag) {
      const { name, value } = datum.filters[i];
      try {
        Filters[name].apply(datum, value);
      } catch (error: any) {
        datum.filters[i] = {
          ...datum.filters[i],
          error: error.message,
        };
      }
    }
  }
  updateData(datum);
}
interface AddFilterOptions {
  filterIndex?: number;
  isDeleteAllow?: boolean;
}

export function addFilter(
  datum: Entry1D | Entry2D,
  filterOptions: FilterOption,
  options: AddFilterOptions = {},
) {
  const { filterIndex, isDeleteAllow = true } = options;
  const id = v4();

  const filter = {
    ...filterOptions,
    id,
    flag: true,
    isDeleteAllow,
  };

  if (typeof filterIndex === 'number' && filterIndex !== -1) {
    datum.filters = [
      ...datum.filters.slice(0, filterIndex + 1),
      filter,
      ...datum.filters.slice(filterIndex + 1),
    ];
  } else {
    datum.filters.push(filter);
  }
}

function replaceFilter(datum: Entry1D | Entry2D, filterID: string, value: any) {
  // datum.filters = datum.filters.slice(0);
  const index = datum.filters.findIndex((f) => f.id === filterID);
  datum.filters[index] = {
    ...datum.filters[index],
    value,
  };
}
