import { NmrData1D } from 'cheminfo-types';

import { Info1D } from '../../../../types/Info/Info1D';
import { FilterDomainUpdateRules } from '../../../FiltersManager';
import { Entry1D } from '../../Entry1D';

import { apodizationFilter } from './apodizationFilter';

export const id = 'apodization';
export const name = 'Apodization';
export const DOMAIN_UPDATE_RULES: Readonly<FilterDomainUpdateRules> = {
  updateXDomain: true,
  updateYDomain: false,
};

export interface ApodizationOptions {
  lineBroadening: number;
  gaussBroadening: number;
  lineBroadeningCenter: number;
}

type Entry1DApodizationFilter = Partial<Omit<Entry1D, 'info'>> & {
  info: Partial<Info1D>;
  data: NmrData1D;
};

export function apply(
  datum1D: Entry1DApodizationFilter,
  options: ApodizationOptions,
) {
  if (!isApplicable(datum1D)) {
    throw new Error('apodization not applicable on this data');
  }

  const {
    lineBroadening = 1,
    gaussBroadening = 0,
    lineBroadeningCenter = 0,
  } = options;

  // if no options is set, we add the default value to filter
  if (Object.keys(options).length === 0) {
    const filter = datum1D.filters?.find((filter) => filter.name === id);
    if (filter) {
      filter.value = { lineBroadening, gaussBroadening, lineBroadeningCenter };
    }
  }
  const newData = apodizationFilter(datum1D, {
    lineBroadening,
    gaussBroadening,
    lineBroadeningCenter,
  });

  datum1D.data = {
    ...datum1D.data,
    re: newData.re as Float64Array,
    im: newData.im as Float64Array,
  };
}
export function isApplicable(
  datum1D: Entry1DApodizationFilter,
): datum1D is Entry1DApodizationFilter & { data: Required<NmrData1D> } {
  if (datum1D.info.isComplex && datum1D.info.isFid) return true;
  return false;
}

export function reduce(previousValue: any, newValue: any) {
  return {
    once: true,
    reduce: newValue,
  };
}
