import airPLS from 'ml-airpls';
import baselineRegression from 'ml-baseline-correction-regression';
import { xyEquallySpaced } from 'ml-spectra-processing';

import { FilterDomainUpdateRules } from '../../FiltersManager';
import { Entry1D } from '../Entry1D';

export const id = 'baselineCorrection';
export const name = 'Baseline correction';
export const DOMAIN_UPDATE_RULES: Readonly<FilterDomainUpdateRules> = {
  updateXDomain: false,
  updateYDomain: true,
};

export type BaselineAlgorithms = 'airpls' | 'polynomial';

export interface BaselineCorrectionZone {
  from: number;
  to: number;
  id: string;
}

export interface PolynomialOptions {
  zones: BaselineCorrectionZone[];
  algorithm: 'polynomial';
  degree: number;
}

export interface AirplsOptions {
  zones: BaselineCorrectionZone[];
  algorithm: 'airpls';
  maxIterations: number;
  tolerance: number;
}

export type BaselineCorrectionOptions = PolynomialOptions | AirplsOptions;

export function apply(datum1D: Entry1D, options: BaselineCorrectionOptions) {
  if (!isApplicable(datum1D)) {
    throw new Error('baselineCorrection not applicable on this data');
  }
  const { algorithm } = options;
  const { x, re } = datum1D.data;

  let corrected;
  switch (algorithm) {
    case 'airpls':
      corrected = airPLS(x, re, options).corrected;
      break;
    case 'polynomial':
      {
        const { degree, zones } = options;
        const reduced = xyEquallySpaced(
          { x, y: re },
          { numberOfPoints: 4096, zones },
        );
        const result = baselineRegression(reduced.x, reduced.y, {
          regressionOptions: degree,
        });
        const { regression } = result;
        corrected = new Float64Array(x.length);
        for (let i = 0; i < re.length; i++) {
          corrected[i] = re[i] - regression.predict(x[i]);
        }
      }

      break;
    default:
      // eslint-disable-next-line @typescript-eslint/restrict-template-expressions
      throw new Error(`baselineCorrection: algorithm unknown: ${algorithm}`);
  }

  Object.assign(datum1D.data, { re: corrected });
}

export function isApplicable(datum1D: Entry1D) {
  if (!datum1D.info.isFid) return true;
  return false;
}

export function reduce(
  previousValue: BaselineCorrectionOptions,
  newValue: BaselineCorrectionOptions,
) {
  return {
    once: true,
    reduce: newValue,
  };
}
