import { NmrData1D } from 'cheminfo-types';
import {
  reimAbsolute,
  reimAutoPhaseCorrection,
  reimPhaseCorrection,
  xEnsureFloat64,
} from 'ml-spectra-processing';

import { FilterDomainUpdateRules } from '../../FiltersManager';
import { Entry1D } from '../Entry1D';

export const id = 'phaseCorrection';
export const name = 'Phase correction';
export const DOMAIN_UPDATE_RULES: Readonly<FilterDomainUpdateRules> = {
  updateXDomain: false,
  updateYDomain: false,
};
export interface PhaseCorrectionParameters {
  ph0: number;
  ph1: number;
  absolute: boolean;
}

/**
 *
 * @param {Datum1d} datum1d
 * @param {Object} [options={}]
 * @param {number} [options.ph0=0]
 * @param {number} [options.ph1=0]
 */

export function apply(datum1D: Entry1D, options: any = {}) {
  if (!isApplicable(datum1D)) {
    throw new Error('phaseCorrection not applicable on this data');
  }

  const { absolute = false } = options;
  const filter = datum1D.filters?.find((filter) => filter.name === id);

  if (absolute) {
    const { re, im } = datum1D.data;
    datum1D.data.re = reimAbsolute({ re, im });
    datum1D.data.im = new Float64Array(0);
    if (filter) {
      filter.value = { ...filter.value, ph0: 0, ph1: 0, absolute };
    }
  } else if ('ph0' in options && 'ph1' in options) {
    const { ph0, ph1 } = options;
    phaseCorrection(datum1D, { ph0, ph1 });
    if (filter) {
      filter.value = { ...filter.value, absolute };
    }
  } else {
    const { ph0, ph1 } = autoPhaseCorrection(datum1D);
    phaseCorrection(datum1D, { ph0, ph1 });
    if (filter) {
      filter.value = { ...filter.value, absolute, ph0, ph1 };
    }
  }
}

function phaseCorrection(
  spectrum: Entry1D & { data: Required<NmrData1D> },
  phaseValues: { ph0: number; ph1: number },
) {
  const { ph0, ph1 } = phaseValues;
  const { re, im } = spectrum.data;
  const phased = reimPhaseCorrection(
    { re, im },
    (ph0 * Math.PI) / 180,
    (ph1 * Math.PI) / 180,
    { reverse: true },
  );
  spectrum.data = {
    ...spectrum.data,
    ...{
      re: xEnsureFloat64(phased.re),
      im: xEnsureFloat64(phased.im),
    },
  };
}

export interface AutoPhaseCorrectionOptions {
  minRegSize?: number;
  maxDistanceToJoin?: number;
  magnitudeMode?: boolean;
  factorNoise?: number;
  reverse?: boolean;
}

function autoPhaseCorrection(
  spectrum: Entry1D,
  options: AutoPhaseCorrectionOptions = {},
) {
  const {
    minRegSize = 5,
    maxDistanceToJoin = 128,
    magnitudeMode = true,
    factorNoise = 5,
    reverse = true,
  } = options;

  return reimAutoPhaseCorrection(spectrum.data as any, {
    minRegSize,
    maxDistanceToJoin,
    magnitudeMode,
    factorNoise,
    reverse,
  });
}

export function isApplicable(
  spectrum: Entry1D,
): spectrum is Entry1D & { data: Required<NmrData1D> } {
  if (spectrum.info.isComplex && !spectrum.info.isFid) return true;
  return false;
}

export function reduce(
  previousValue: PhaseCorrectionParameters,
  newValue: PhaseCorrectionParameters,
) {
  return {
    once: true,
    reduce: newValue,
  };
}
