import { v4 as generateID } from '@lukeed/uuid';
import omit from 'lodash/omit';

import { DatumKind } from '../../../constants/SignalsKinds';
import { Range } from '../../../types/Ranges/Ranges';
import { Signal1D } from '../../../types/Signals/Signal1D';

export function getDiaIDs(range: Range): string[] {
  return ([] as string[]).concat(
    range.diaIDs || [],
    range.signals
      ? range.signals.flatMap((_signal) => _signal.diaIDs || [])
      : [],
  );
}

export function getNbAtoms(range: Range, signalIndex?: number): number {
  if (signalIndex === undefined) {
    let sum = 0;
    if (range.signals) {
      for (const signal of range.signals) {
        sum += signal.nbAtoms || 0;
      }
    }
    return sum;
  }

  return range.signals[signalIndex].nbAtoms || 0;
}

export function setNbAtoms(range: Range, signalIndex?: number): void {
  range.nbAtoms = getNbAtoms(range, signalIndex);
  if (range.nbAtoms === 0) {
    delete range.nbAtoms;
  }
}

export function resetDiaIDs(range: Range): void {
  const deletedKeys = ['diaIDs', 'nbAtoms'];
  range = omit(range, deletedKeys) as Range;
  range.signals = range.signals.map((signal) => {
    return omit(signal, deletedKeys) as Signal1D;
  });
}

/**
 *
 * @param {object} range
 * @param {String('both' | 'range' | 'signal')} unlinkType
 * @param {object} options
 * @param {number} [options.signalIndex]
 * @returns {object}
 */

interface UnlinkOptions {
  unlinkType?: 'both' | 'range' | 'signal';
  signalIndex?: number;
}
export function unlink(range: Range, options: UnlinkOptions = {}): Range {
  const { unlinkType = 'both', signalIndex = -1 } = options;
  const keys = ['diaIDs', 'nbAtoms'];

  switch (unlinkType) {
    case 'both':
      resetDiaIDs(range);
      break;
    case 'range':
      range = omit(range, keys) as Range;
      break;
    case 'signal':
      if (signalIndex !== -1) {
        range.signals[signalIndex] = omit(
          range.signals[signalIndex],
          keys,
        ) as Signal1D;
      }
      break;
    default:
      break;
  }
  return range;
}

export function addDefaultSignal(range: Range) {
  range.signals.push({
    id: generateID(),
    multiplicity: 'm',
    kind: 'signal',
    js: [],
    delta: (range.to + range.from) / 2,
  });
}

export function checkRangeKind(range: Range) {
  return range.kind === DatumKind.signal;
}

export function checkSignalKinds(range: Range, kinds: string[]) {
  return !range.signals.some(
    (_signal) => _signal.kind === undefined || !kinds.includes(_signal.kind),
  );
}
