import { generateSpectrum, OptionsSG1D } from 'spectrum-generator';

import { NMRPeak1D } from './NMRPeak1D';
import { convertWidthToPPM } from './util/convertWidthToPPM';

/**
 * Convert an array of peaks assuming that width/fwhm is in Hz
 */

export interface PeaksToXYOptions extends OptionsSG1D {
  /**
   * First x value (inclusive).
   * @default `firstPeak.x + firstPeak.width * 3`
   */
  from?: number;
  /**
   * Last x value (inclusive).
   * @default `lastPeak.x + lastPeak.width * 3`
   */
  to?: number;
  /**
   * frequency of the spectrum
   */
  frequency: number;
}

export function peaksToXY(peaks: NMRPeak1D[], options: PeaksToXYOptions) {
  const { frequency, nbPoints = 1024, shape } = options;

  if (!frequency) {
    throw new Error('frequency is mandatory');
  }
  const newPeaks = convertWidthToPPM(peaks, { frequency });
  return generateSpectrum(newPeaks, {
    generator: {
      ...getFromTo(newPeaks, options),
      nbPoints,
      shape,
    },
  });
}

function getFromTo(newPeaks: NMRPeak1D[], options: PeaksToXYOptions) {
  if ('to' in options && 'from' in options) {
    return {
      from: options.from,
      to: options.to,
    };
  }

  newPeaks.sort((a, b) => a.x - b.x);
  const firstPeak = newPeaks[0];
  const lastPeak = newPeaks[newPeaks.length - 1];

  const {
    from = firstPeak.x - (firstPeak.width * 2) / options.frequency,
    to = lastPeak.x + (lastPeak.width * 2) / options.frequency,
  } = options;

  return {
    from,
    to,
  };
}
