import { readFileSync } from 'node:fs';
import { join } from 'path';

import { FifoLogger } from 'fifo-logger';
import OCL, { Molecule } from 'openchemlib/minimal';

import type { NMRSignal2D } from '../../xyz/NMRSignal2D';
import { predictAll } from '../predictAll';

import carbonDB from './data/carbonDB-withoutStatistic.json';

const molfile = `Benzene, ethyl-, ID: C100414
  NIST    16081116462D 1   1.00000     0.00000
Copyright by the U.S. Sec. Commerce on behalf of U.S.A. All rights reserved.
  8  8  0     0  0              1 V2000
    0.5015    0.0000    0.0000 C   0  0  0  0  0  0           0  0  0
    0.0000    0.8526    0.0000 C   0  0  0  0  0  0           0  0  0
    1.5046    0.0000    0.0000 C   0  0  0  0  0  0           0  0  0
    2.0062    0.8526    0.0000 C   0  0  0  0  0  0           0  0  0
    3.0092    0.8526    0.0000 C   0  0  0  0  0  0           0  0  0
    1.5046    1.7554    0.0000 C   0  0  0  0  0  0           0  0  0
    0.5015    1.7052    0.0000 C   0  0  0  0  0  0           0  0  0
    3.5108    0.0000    0.0000 C   0  0  0  0  0  0           0  0  0
  1  2  2  0     0  0
  3  1  1  0     0  0
  2  7  1  0     0  0
  4  3  2  0     0  0
  4  5  1  0     0  0
  6  4  1  0     0  0
  5  8  1  0     0  0
  7  6  2  0     0  0
M  END
`;

describe('predict all nmr spectrum', () => {
  const molecule = OCL.Molecule.fromMolfile(molfile);

  it.skip('cyclosporin', async () => {
    const logger = new FifoLogger();
    const molfile = readFileSync(
      join(__dirname, 'data/cyclosporin.mol'),
      'utf-8',
    );
    const molecule = Molecule.fromMolfile(molfile);
    const result = await predictAll(molecule, {
      predictOptions: {
        C: {
          database: carbonDB,
        },
      },
    });
    const spectra = result.spectra;
    expect(logger.getLogs()).toHaveLength(2);
    expect(Object.keys(spectra)).toHaveLength(1);
    expect(spectra.carbon.signals).toHaveLength(6);
    expect(spectra.carbon.joinedSignals).toHaveLength(3);
  });

  it('hmbc benzene', async () => {
    jest.setTimeout(50000);
    const result = await predictAll(molecule, {
      predictOptions: {
        C: {
          database: carbonDB,
        },
      },
    });
    const signals = result.spectra.hmbc.joinedSignals;

    const expectedDeltasHMBC = [
      [0.992, 28.7],
      [0.992, 144.2],
      [2.653, 14.6],
      [2.653, 144.2],
      [2.653, 128.8],
      [7.162, 28.7],
      [7.162, 144.2],
      [7.162, 128.4],
      [7.162, 127.8],
      [7.162, 128.8],
      [7.26, 144.2],
      [7.26, 128.8],
      [7.26, 127.8],
      [7.26, 128.4],
      [7.196, 128.8],
      [7.196, 128.4],
    ];

    const stringDeltas = expectedDeltasHMBC.map((d) =>
      d.map((e) => e.toFixed(3)).join('-'),
    );
    const deltas = signals.map(
      (signal: NMRSignal2D) =>
        `${signal.x.delta.toFixed(3)}-${signal.y.delta.toFixed(3)}`,
    );
    stringDeltas.sort((a: string, b: string) => a.localeCompare(b));
    deltas.sort((a: string, b: string) => a.localeCompare(b));
    expect(stringDeltas).toStrictEqual(deltas);

    const cosySignals = result.spectra.cosy.joinedSignals;
    const cosyDeltas = cosySignals.map((e: NMRSignal2D) => [
      e.x.delta,
      e.y.delta,
    ]);

    expect(cosyDeltas).toStrictEqual([
      [7.26, 7.196],
      [7.26, 7.162],
      [7.196, 7.26],
      [7.162, 7.26],
      [2.653, 0.992],
      [0.992, 2.653],
      [7.26, 7.26],
      [7.196, 7.196],
      [7.162, 7.162],
      [2.653, 2.653],
      [0.992, 0.992],
    ]);

    const hsqcSignals = result.spectra.hsqc.joinedSignals;
    const hsqcDeltas = hsqcSignals.map((e: NMRSignal2D) => [
      e.x.delta,
      e.y.delta,
    ]);

    expect(hsqcDeltas).toStrictEqual([
      [7.26, 128.4],
      [7.196, 127.8],
      [7.162, 128.8],
      [2.653, 28.7],
      [0.992, 14.6],
    ]);
  }, 30000);

  it('webservice', async () => {
    const result = await predictAll(molecule, {
      predictOptions: {
        C: {
          webserviceURL:
            'https://nmr-prediction.service.zakodium.com/v1/predict/carbon',
        },
      },
    });
    const signals = result.spectra.hmbc.joinedSignals;

    const expectedDeltasHMBC = [
      [0.992, 28.7],
      [0.992, 144.2],
      [2.653, 14.6],
      [2.653, 144.2],
      [2.653, 128.8],
      [7.162, 28.7],
      [7.162, 144.2],
      [7.162, 128.4],
      [7.162, 127.8],
      [7.162, 128.8],
      [7.26, 144.2],
      [7.26, 128.8],
      [7.26, 127.8],
      [7.26, 128.4],
      [7.196, 128.8],
      [7.196, 128.4],
    ];

    const stringDeltas = expectedDeltasHMBC.map((d) =>
      d.map((e) => e.toFixed(3)).join('-'),
    );
    const deltas = signals.map(
      (signal: NMRSignal2D) =>
        `${signal.x.delta.toFixed(3)}-${signal.y.delta.toFixed(3)}`,
    );
    stringDeltas.sort((a: string, b: string) => a.localeCompare(b));
    deltas.sort((a: string, b: string) => a.localeCompare(b));
    expect(stringDeltas).toStrictEqual(deltas);

    const cosySignals = result.spectra.cosy.joinedSignals;
    const cosyDeltas = cosySignals.map((e: NMRSignal2D) => [
      e.x.delta,
      e.y.delta,
    ]);

    expect(cosyDeltas).toStrictEqual([
      [7.26, 7.196],
      [7.26, 7.162],
      [7.196, 7.26],
      [7.162, 7.26],
      [2.653, 0.992],
      [0.992, 2.653],
      [7.26, 7.26],
      [7.196, 7.196],
      [7.162, 7.162],
      [2.653, 2.653],
      [0.992, 0.992],
    ]);

    const hsqcSignals = result.spectra.hsqc.joinedSignals;
    const hsqcDeltas = hsqcSignals.map((e: NMRSignal2D) => [
      e.x.delta,
      e.y.delta,
    ]);

    expect(hsqcDeltas).toStrictEqual([
      [7.26, 128.4],
      [7.196, 127.8],
      [7.162, 128.8],
      [2.653, 28.7],
      [0.992, 14.6],
    ]);
  });
});

describe('No non-labile hydrogens', () => {
  it('prediction', async () => {
    const logger = new FifoLogger();
    const molfile = readFileSync(join(__dirname, 'data/bromo.mol'), 'utf-8');
    const molecule = Molecule.fromMolfile(molfile);
    molecule.addImplicitHydrogens();
    const result = await predictAll(molecule, { logger });
    const spectra = result.spectra;
    expect(logger.getLogs()).toHaveLength(2);
    expect(Object.keys(spectra)).toHaveLength(1);
    expect(spectra.carbon.signals).toHaveLength(6);
    expect(spectra.carbon.joinedSignals).toHaveLength(3);
  });
});
