import { existsSync, mkdirSync, readFileSync, writeFileSync } from 'fs';
import { join } from 'path';

//@ts-expect-error there is not a type declaration file
import md5 from 'md5';
import OCL from 'openchemlib/minimal';

import type { NMRSignal1D } from '../../signals/NMRSignal1D';
import type { MakeMandatory } from '../../utilities/MakeMandatory';
import type { NMRRange } from '../../xy/NMRRange';
import { predictProton } from '../predictProton';

const molfile = `Benzene, ethyl-, ID: C100414
  NIST    16081116462D 1   1.00000     0.00000
Copyright by the U.S. Sec. Commerce on behalf of U.S.A. All rights reserved.
  8  8  0     0  0              1 V2000
    0.5015    0.0000    0.0000 C   0  0  0  0  0  0           0  0  0
    0.0000    0.8526    0.0000 C   0  0  0  0  0  0           0  0  0
    1.5046    0.0000    0.0000 C   0  0  0  0  0  0           0  0  0
    2.0062    0.8526    0.0000 C   0  0  0  0  0  0           0  0  0
    3.0092    0.8526    0.0000 C   0  0  0  0  0  0           0  0  0
    1.5046    1.7554    0.0000 C   0  0  0  0  0  0           0  0  0
    0.5015    1.7052    0.0000 C   0  0  0  0  0  0           0  0  0
    3.5108    0.0000    0.0000 C   0  0  0  0  0  0           0  0  0
  1  2  2  0     0  0
  3  1  1  0     0  0
  2  7  1  0     0  0
  4  3  2  0     0  0
  4  5  1  0     0  0
  6  4  1  0     0  0
  5  8  1  0     0  0
  7  6  2  0     0  0
M  END
`;

const cache = (molfile: string, value: any) => {
  const hash = md5(molfile);
  const cacheDir = join(__dirname, 'cache');
  const cacheFilename = join(cacheDir, hash);
  if (!existsSync(cacheDir)) mkdirSync(cacheDir);
  if (value === undefined) {
    if (existsSync(cacheFilename)) {
      return readFileSync(cacheFilename, 'utf8');
    }
  } else {
    writeFileSync(cacheFilename, value, 'utf8');
  }
};

type Signal1DWithJs = MakeMandatory<NMRSignal1D, 'js'>;
type RangeWithSignals = Omit<NMRRange, 'signals'> & {
  signals: Signal1DWithJs[];
};
function checkRange(range: NMRRange): asserts range is RangeWithSignals {
  if (!range.signals) throw new Error('There is not signals in range');
  for (const signal of range.signals) {
    if (!signal.js) throw new Error('There is not js in signal');
  }
}
describe('predictProton', () => {
  it('1H chemical shift prediction', async () => {
    const molecule = OCL.Molecule.fromMolfile(molfile);
    const prediction = await predictProton(molecule, { cache });
    expect(Object.keys(prediction)).toStrictEqual([
      'molfile',
      'diaIDs',
      'nucleus',
      'joinedSignals',
      'signals',
      'ranges',
      'molecule',
    ]);
    const firstSignal = prediction.signals[0];

    expect(firstSignal).toStrictEqual({
      id: firstSignal.id,
      atoms: [8],
      diaIDs: ['did@`@f\\bbRaih@J@A~dHBIU@'],
      nbAtoms: 1,
      delta: 7.26,
      js: [
        {
          coupling: 7.758,
          atoms: [10],
          diaIDs: ['did@`@fTfYUn`HH@GzP`HeT'],
          multiplicity: 'd',
          pathLength: 3,
        },
        {
          coupling: 7.718,
          atoms: [9],
          diaIDs: ['did@`@fTfUvf`@h@GzP`HeT'],
          multiplicity: 'd',
          pathLength: 3,
        },
        {
          coupling: 1.596,
          atoms: [14],
          diaIDs: ['did@`@f\\bbRaih@J@A~dHBIU@'],
          multiplicity: 'd',
          pathLength: 4,
        },
        {
          coupling: 0.507,
          atoms: [13],
          diaIDs: ['did@`@fTfYUn`HH@GzP`HeT'],
          multiplicity: 'd',
          pathLength: 5,
        },
      ],
    });
    const firstJoinedSignal = prediction.joinedSignals[0];
    expect(firstJoinedSignal).toStrictEqual({
      id: firstJoinedSignal.id,
      nbAtoms: 3,
      delta: 0.992,
      multiplicity: 't',
      diaIDs: ['did@`@fTeYWaj@@@GzP`HeT'],
      atoms: [15, 16, 17],
      js: [
        {
          coupling: 7.392,
          multiplicity: 't',
          diaIDs: ['did@`@fTf[Waj@@bJ@_iB@bUP'],
          pathLength: 3,
        },
      ],
    });
    expect(prediction.ranges).toHaveLength(3);
    const lastRange = prediction.ranges[2];
    checkRange(lastRange);
    expect(lastRange.integration).toBe(5);
    expect(lastRange.signals).toHaveLength(3);
    expect(lastRange.signals[0].js).toHaveLength(4);
  }, 30000);
  it('empty molecule', async () => {
    const molecule = new OCL.Molecule(16, 16);
    const result = await predictProton(molecule);
    expect(result.diaIDs).toStrictEqual([]);
    expect(result.joinedSignals).toStrictEqual([]);
    expect(result.signals).toStrictEqual([]);
    expect(result.ranges).toStrictEqual([]);
    expect(result.molecule.getAllAtoms()).toBe(0);
    expect(result.molecule.getAllBonds()).toBe(0);
  }, 30000);
});
