import type { Logger } from 'cheminfo-types';
import type { Molecule } from 'openchemlib/full';
import { getDiastereotopicAtomIDs } from 'openchemlib-utils';

import type { Prediction1D } from './Prediction1D';
import { predictCOSY } from './predictCOSY';
import { predictCarbon, PredictCarbonOptions } from './predictCarbon';
import { predictHMBC } from './predictHMBC';
import { predictHSQC } from './predictHSQC';
import { predictProton, PredictProtonOptions } from './predictProton';
import { getPredictions } from './utils/getPredictions';
import type { Predictor } from './utils/predict2D';

export interface PredictAllOptions {
  /**
   * maximum number of bonds to take into account.
   * @default 1
   */
  maxLength?: number;
  /**
   * minimum number of bonds to take into account
   * @default 1
   */
  minLength?: number;
  /**
   * mono dimensional predictors. The object should has properties with name
   * equal to "H" and "C".
   */
  predictor?: { H?: Predictor; C?: Predictor };
  /**
   *  mono dimensional signal to generate the bidimenionals cross peaks.
   * The object should has properties with name equal to "H" and "C".
   */
  predictOptions?: { H?: PredictProtonOptions; C?: PredictCarbonOptions };
  /**
   * signals with euclidean distance in ppm less than this it will
   * joined into a zone.
   * @default {H:0.05,C:0.5}
   */
  joinDistance?: { H: number; C: number };
  /**
   * mono dimensional signal to generate the bidimenionals cross peaks.
   * The object should has properties with name equal to "H" and "C".
   * the value of each prediction should be similar that predictProton and predictCarbon.
   */
  predictions?: {
    [key: string]: Prediction1D | undefined;
    H?: Prediction1D;
    C?: Prediction1D;
  };
  logger?: Logger;
}

export interface PredictedAll {
  molfile: string;
  diaIDs: string[];
  spectra: Record<string, any>;
}
/**
 * Generate the correlation bidimensional nmr peaks based on the number of bonds between a pair of atoms
 * @returns {Promise<object>} - object with molfile, diaIDs, 1D and 2D signals, joined signals, ranges and zones.
 */
export async function predictAll(
  molecule: Molecule,
  options: PredictAllOptions = {},
): Promise<PredictedAll> {
  const {
    logger,
    predictor = { H: predictProton, C: predictCarbon },
    joinDistance = { H: 0.05, C: 0.5 },
    predictOptions = {},
  } = options;

  const diaIDs = getDiastereotopicAtomIDs(molecule);
  let { predictions } = options;

  const protonPrediction = await getPredictions(
    'H',
    molecule,
    predictOptions,
    predictor,
    predictions,
  );

  const carbonPrediction = await getPredictions(
    'C',
    molecule,
    predictOptions,
    predictor,
    predictions,
  );

  if (!protonPrediction || !carbonPrediction) {
    throw new Error('predictions are not availaible');
  }

  const spectra: any = {};
  if (check1DPrediction(protonPrediction)) {
    spectra.proton = protonPrediction;
  }
  if (check1DPrediction(carbonPrediction)) {
    spectra.carbon = carbonPrediction;
  }

  const { molfile } = protonPrediction;

  predictions = {
    H: protonPrediction,
    C: carbonPrediction,
  };

  if (spectra.proton) {
    spectra.cosy = await predictCOSY(molecule, {
      predictions,
      joinDistance,
    });
  } else if (logger) {
    logger.warn('Proton prediction is missing so COSY could not be generated.');
  }

  if (spectra.carbon && spectra.proton) {
    spectra.hsqc = await predictHSQC(molecule, {
      predictions,
      joinDistance,
    });
    spectra.hmbc = await predictHMBC(molecule, {
      predictions,
      joinDistance,
    });
  } else if (logger) {
    logger.warn(
      'Carbon and / or proton prediction is missing so HSQC and HMBC could not be generated.',
    );
  }

  for (const key in spectra) {
    delete spectra[key].molfile;
    delete spectra[key].diaIDs;
  }

  return {
    molfile,
    diaIDs,
    spectra,
  };
}

function check1DPrediction(prediction?: Prediction1D) {
  if (!prediction) return false;
  if (prediction.ranges.length === 0) return false;
  const { signals, diaIDs } = prediction;
  for (const signal of signals) {
    const isCorrect = signal.diaIDs?.every((diaID) => diaIDs.includes(diaID));
    if (!isCorrect) return false;
  }
  return true;
}
