import { FromTo, PointXY } from 'cheminfo-types';
import type { Molecule } from 'openchemlib/full';

import { signals2DToZ } from '../signals/signals2DToZ';
import { OptionsSignalsToXY, signalsToXY } from '../signals/signalsToXY';
import { getFrequency } from '../utilities/getFrequency';

import { PredictAllOptions, predictAll } from './predictAll';
/**
 * This function will generate an object compatible with NMR-ium
 *
 * @param molecule
 * @param options
 */

interface OneDOptions {
  proton: FromTo;
  carbon: FromTo;
  nbPoints: number;
  lineWidth: number;
}

interface TwoDOptions {
  from: PointXY;
  to: PointXY;
  nbPoints: PointXY;
}
interface SimulationOptions {
  frequency: number;
  oneD?: Partial<OneDOptions>;
  twoD?: Partial<TwoDOptions>;
}
interface PredictAllSpectraOptions {
  prediction?: PredictAllOptions;
  simulation?: SimulationOptions;
}

export async function predictAllSpectra(
  molecule: Molecule,
  options: PredictAllSpectraOptions = {},
) {
  const {
    simulation: simulationOptions = { oneD: {}, twoD: {} } as SimulationOptions,
    prediction: predictionOptions = {},
  } = options;

  const predictions = await predictAll(molecule, predictionOptions);
  const { molfile, spectra } = predictions;

  const oneDOptions: OneDOptions = {
    ...{
      proton: { from: 0, to: 14 },
      carbon: { from: 0, to: 200 },
      nbPoints: 65536,
      lineWidth: 1,
    },
    ...(simulationOptions.oneD || {}),
  };

  const twoDOptions: TwoDOptions = {
    ...{
      from: { x: oneDOptions.proton.from, y: oneDOptions.carbon.from },
      to: { x: oneDOptions.proton.to, y: oneDOptions.carbon.to },
      nbPoints: { x: 1024, y: 1024 },
    },
    ...(simulationOptions.twoD || {}),
  };

  const newSpectra: any[] = [];
  for (const experiment in spectra) {
    switch (experiment) {
      case 'carbon':
      case 'proton': {
        newSpectra.push(
          get1DSpectrum(spectra[experiment], {
            nbPoints: oneDOptions.nbPoints,
            lineWidth: oneDOptions.lineWidth,
            ...oneDOptions[experiment],
            experiment,
            frequency: simulationOptions.frequency,
          }),
        );
        break;
      }
      case 'cosy':
      case 'hsqc':
      case 'hmbc': {
        newSpectra.push(
          get2DSpectrum(spectra[experiment], {
            ...twoDOptions,
            experiment,
            frequency: simulationOptions.frequency,
          }),
        );
        break;
      }
      default:
        break;
    }
  }

  return { spectra: newSpectra, molecules: [{ molfile }] };
}

type Get2DSpectrumOptions = TwoDOptions & {
  experiment: string;
  frequency: number;
};

function get2DSpectrum(prediction: any, options: Get2DSpectrumOptions) {
  const { signals, zones, nuclei } = prediction;
  const { frequency: baseFrequency, experiment } = options;

  const width = get2DWidth(nuclei);
  const frequency = calculateFrequency(nuclei, baseFrequency);

  const minMaxContent = signals2DToZ(signals, {
    ...options,
    width,
    factor: 3,
  });

  const spectrum = {
    data: { rr: { ...minMaxContent, noise: 0.01 } },
    info: {
      nucleus: nuclei,
      originFrequency: frequency,
      baseFrequency: frequency,
      pulseSequence: experiment,
      experiment: '2d',
    },
    zones: { values: zones },
  };
  return spectrum;
}

function get2DWidth(nucleus: string[]) {
  return nucleus[0] === nucleus[1] ? 0.02 : { x: 0.02, y: 0.2133 };
}

type Get1DSpectrumOptions = OptionsSignalsToXY & { experiment: string };

function get1DSpectrum(prediction: any, options: Get1DSpectrumOptions) {
  const { frequency = 400, experiment } = options;
  const { signals, nucleus } = prediction;
  const realFrequency = calculateFrequency(nucleus, frequency) as number;
  const { x, y } = signalsToXY(signals, {
    ...options,
    frequency: realFrequency,
  });
  const spectrum = {
    // todo Array.from is temporary for the developement
    data: { x: Array.from(x), re: Array.from(y), im: null },
    info: {
      nucleus: prediction.nucleus,
      dimension: 1,
      isComplex: false,
      originFrequency: realFrequency,
      baseFrequency: realFrequency,
      pulseSequence: 'prediction',
      experiment,
      isFt: true,
    },
    ranges: { values: prediction.ranges },
  };

  return spectrum;
}

function calculateFrequency(
  nucleus: string | string[],
  frequency: number,
): number | string {
  if (typeof nucleus === 'string') {
    return getFrequency(nucleus, { nucleus: '1H', frequency });
  } else if (nucleus[0] === nucleus[1]) {
    return `${frequency},${frequency}`;
  } else {
    return `${frequency},${getFrequency(nucleus[1], {
      nucleus: nucleus[0],
      frequency,
    })}`;
  }
}
