import { v4 as generateID } from '@lukeed/uuid';
import type { Molecule } from 'openchemlib/full';
import { getDiastereotopicAtomIDs } from 'openchemlib-utils';

import { setIDs } from '../peaks/util/setIDs';
import type { NMRSignal1D } from '../signals/NMRSignal1D';
import { signalsToRanges } from '../signals/signalsToRanges';
import type { MakeMandatory } from '../utilities/MakeMandatory';

import type { Prediction1D } from './Prediction1D';
import type { DataBaseStructure } from './dataStructure';
import { fetchPrediction } from './utils/fetchPrediction';
import { getFilteredIDiaIDs } from './utils/getFilteredIDiaIDs';
import { isEmptyMolecule } from './utils/isEmptyMolecule';
import { Prediction, queryByHose } from './utils/queryByHOSE';

export interface PredictCarbonOptions {
  /**
   *  custom database, each entry in the levels should has
   * an array as value [median] or [median, mean, sd, min, max, nb] for statistic purpose.
   */
  database?: DataBaseStructure;
  /**
   * URL of a webService that returns the prediction.
   */
  webserviceURL?: string;
  /**
   *  max level to take into account in the query. If is not specified
   * the max level in the database will be used.
   * @default 'maximum level in the database'
   */
  maxSphereSize?: number;
  /**
   * diastereotopic atom ids.
   */
  diaIDs?: string[];
}

export type PredictCarbon = typeof predictCarbon;

type Signal1DFromPrediction = MakeMandatory<
  NMRSignal1D,
  'nbAtoms' | 'atoms' | 'diaIDs'
>;
function checkFromPrediction(
  signal: NMRSignal1D,
): asserts signal is Signal1DFromPrediction {
  if (!signal.atoms) throw new Error('There is not atoms');
  if (!signal.diaIDs) throw new Error('There is not diaIDs');
  if (!signal.nbAtoms) throw new Error('There is not nbAtoms');
}
/**
 * Make a query to a hose code based database to predict carbon chemical shift
 * @returns {Promise<object>} - object with molfile, diaIDs, signals, joined signals by diaIDs and ranges.
 */

export async function predictCarbon(
  molecule: Molecule,
  options: PredictCarbonOptions = {},
): Promise<Prediction1D> {
  if (isEmptyMolecule(molecule)) {
    return {
      molfile: molecule.toMolfile(),
      nucleus: '13C',
      diaIDs: [],
      joinedSignals: [],
      signals: [],
      ranges: [],
      molecule,
    };
  }

  const {
    database,
    webserviceURL = 'https://nmr-prediction.service.zakodium.com/v1/predict/carbon',
    diaIDs = getDiastereotopicAtomIDs(molecule),
  } = options;

  if (!database) return fetchPrediction(molecule, { webserviceURL });

  if (!database) {
    throw new Error('There is not a database');
  }

  const maxLevel = database.length - 1;

  let { maxSphereSize = maxLevel } = options;

  if (maxSphereSize > maxLevel) maxSphereSize = maxLevel;

  const { carbonDiaIDs, molfile } = getFilteredIDiaIDs(molecule, {
    maxSphereSize,
    diaIDs,
  });

  const predictions = queryByHose(carbonDiaIDs, database, {
    maxSphereSize,
  });

  const signals = formatSignals(predictions);
  const joinedSignals = joinSignalByDiaID(signals);
  return {
    molfile,
    nucleus: '13C',
    diaIDs,
    joinedSignals,
    signals,
    ranges: signalsToRanges(joinedSignals),
    molecule,
  };
}

function formatSignals(predictions: Prediction[]) {
  const signals = [];
  for (const prediction of predictions) {
    const { atoms, nbAtoms, delta, diaIDs, statistic } = prediction;
    const signal = {
      id: generateID(),
      delta: delta || NaN,
      atoms,
      diaIDs,
      multiplicity: 's',
      nbAtoms,
      statistic,
      js: [],
    };
    signals.push(signal as NMRSignal1D);
  }
  return signals;
}

function joinSignalByDiaID(signals: NMRSignal1D[]) {
  const joinedSignals: Record<string, Signal1DFromPrediction> = {};
  for (const signal of signals) {
    checkFromPrediction(signal);
    const diaID = signal.diaIDs[0];
    if (!joinedSignals[diaID]) {
      joinedSignals[diaID] = JSON.parse(
        JSON.stringify(signal),
      ) as Signal1DFromPrediction;
    } else {
      joinedSignals[diaID].nbAtoms += signal.nbAtoms;
      joinedSignals[diaID].atoms.push(...signal.atoms);
    }
  }
  return setIDs(Object.values(joinedSignals));
}
