import { toBeDeepCloseTo } from 'jest-matcher-deep-close-to';

import { hackSignalsToXY } from '../hackSignalsToXY';
import { optimizeSignals } from '../optimizeSignals';

expect.extend({ toBeDeepCloseTo });

describe('simple optimization', () => {
  it('one doublet', () => {
    const frequency = 100;
    const nbPoints = 1024;
    const delta = 0.5;
    const coupling = 8;
    const atoms = [3];
    const lineWidth = 2;
    const maxValue = 2;
    const multiplicity = 'd';
    const [from, to] = [delta - 0.1, delta + 0.1];
    const doubletData = hackSignalsToXY(
      [
        {
          delta,
          atoms,
          js: [{ multiplicity, coupling }],
        },
      ],
      { frequency, nbPoints, from, to, lineWidth, maxValue },
    );
    const newSignal = optimizeSignals(
      doubletData,
      [
        {
          delta: 0.51,
          js: [{ multiplicity, coupling: coupling + 0.1 }],
        },
      ],
      {
        lineWidth: 2,
        optimization: {
          kind: 'lm',
          options: {
            maxIterations: 20,
          },
        },
        simulation: {
          frequency,
          maxClusterSize: 1,
        },
      },
    )[0];
    expect(newSignal.delta).toBeDeepCloseTo(delta, 2);
    expect(newSignal.intensity).toBeDeepCloseTo(maxValue, 2);
    for (const peak of newSignal.peaks) {
      const { shape, x, y } = peak;
      expect(Math.abs(x - delta)).toBeDeepCloseTo(
        Math.abs(coupling / frequency / 2),
        2,
      );
      expect(y).toBeDeepCloseTo(maxValue, 2);
      expect(shape.fwhm).toBeDeepCloseTo(lineWidth);
    }
  });
  it('two overlaped doublet', () => {
    const frequency = 100;
    const nbPoints = 1024;
    const deltas = [0.49, 0.51, 0.51];
    const coupling = [2, 2, 2];
    const lineWidth = 0.5;
    const maxValue = 3;
    const multiplicity = ['d', 'd', 'd'];
    const [from, to] = [0.4, 0.6];
    const signals = deltas.map((delta, i) => {
      return {
        delta,
        js: [{ multiplicity: multiplicity[i], coupling: coupling[i] }],
      };
    });

    const doubletData = hackSignalsToXY(signals, {
      frequency,
      nbPoints,
      from,
      to,
      lineWidth,
      maxValue,
    });
    deltas.splice(2, 1);
    const newSignals = optimizeSignals(
      doubletData,
      deltas.map((delta, i) => {
        return {
          delta: delta + 0.001,
          js: [
            {
              multiplicity: multiplicity[i],
              coupling: coupling[i] + 0.01,
            },
          ],
        };
      }),
      {
        lineWidth: lineWidth + 0.01,
        optimization: {
          kind: 'lm',
          options: {
            maxIterations: 30,
          },
        },
        simulation: {
          frequency,
          maxClusterSize: 1,
        },
      },
    );
    for (let i = 0; i < newSignals.length; i++) {
      expect(newSignals[i].delta).toBeDeepCloseTo(deltas[i], 2);
      expect(newSignals[i].intensity).toBeDeepCloseTo(i + 1);
      for (const peak of newSignals[i].peaks) {
        const { shape, x } = peak;
        expect(Math.abs(x - deltas[i])).toBeDeepCloseTo(
          Math.abs(coupling[i] / frequency / 2),
          2,
        );
        expect(shape.fwhm).toBeDeepCloseTo(lineWidth, 2);
      }
    }
  });
  it('two overlaped doublet. get a guest with direct', () => {
    const frequency = 100;
    const nbPoints = 1024;
    const deltas = [0.49, 0.51];
    const coupling = [2, 2];
    const lineWidth = 1;
    const maxValue = 3;
    const multiplicity = ['d', 'd'];
    const [from, to] = [0.4, 0.6];
    const signals = deltas.map((delta, i) => {
      return {
        delta,
        js: [{ multiplicity: multiplicity[i], coupling: coupling[i] }],
      };
    });
    signals.push(JSON.parse(JSON.stringify(signals[1])));
    const doubletData = hackSignalsToXY(signals, {
      frequency,
      nbPoints,
      from,
      to,
      lineWidth,
      maxValue,
    });

    const tempSignals = optimizeSignals(
      doubletData,
      deltas.map((delta, i) => {
        return {
          delta: 5,
          intensity: 1,
          js: [
            {
              multiplicity: multiplicity[i],
              coupling: 2.5,
            },
          ],
          parameters: {
            delta: {
              min: 0.485,
              max: 0.515,
            },
            fwhm: {
              min: 0.8,
              max: 1.3,
            },
            intensity: {
              min: 0,
              max: maxValue,
            },
          },
        };
      }),
      {
        baseline: 0,
        optimization: {
          kind: 'direct',
          options: {
            iterations: 5,
          },
        },
        simulation: {
          frequency,
          maxClusterSize: 1,
        },
      },
    );

    const newSignals = optimizeSignals(doubletData, tempSignals, {
      baseline: 0,
      optimization: {
        kind: 'lm',
        options: {
          maxIterations: 25,
        },
      },
      simulation: {
        frequency,
        maxClusterSize: 1,
      },
    });
    newSignals.sort((a, b) => a.delta - b.delta);
    for (let i = 0; i < newSignals.length; i++) {
      expect(newSignals[i].delta).toBeDeepCloseTo(deltas[i], 2);
      expect(newSignals[i].intensity).toBeDeepCloseTo(i + 1);
      for (const peak of newSignals[i].peaks) {
        const { shape, x } = peak;
        expect(Math.abs(x - deltas[i])).toBeDeepCloseTo(
          Math.abs(coupling[i] / frequency / 2),
          2,
        );
        expect(shape.fwhm).toBeDeepCloseTo(lineWidth, 2);
      }
    }
  });
});
