import { DataXY } from 'cheminfo-types';
import { getShape1D, Shape1D, Shape1DInstance } from 'ml-peak-shape-generator';
import { xyFindClosestPoint } from 'ml-spectra-processing';

import { MakeMandatory } from '../../utilities/MakeMandatory';
import { Jcoupling } from '../Jcoupling';
import { Signal } from '../optimizeSignals';

import { defaultParameters } from './defaultParameters';
import { signalsToPointXY } from './signalsToPointXY';

const properties = ['init', 'min', 'max', 'gradientDifference'];

export interface InternalSignal {
  shape: Shape1D;
  shapeFct: Shape1DInstance;
  signal: MakeMandatory<Signal, 'js' | 'intensity'>;
  parameters: string[];
  propertiesValues: Record<string, number[]>;
  fromIndex: number;
  toIndex: number;
  fromIndexCoupling: number;
  toIndexCoupling: number;
}
export type Parameter = 'delta' | 'intensity' | 'fwhm' | 'mu' | 'coupling';

export type ParametersFromOptions = Record<
  Parameter,
  Record<
    string,
    | number
    | ((options?: {
        signal?: Signal;
        shape?: Shape1D;
        index?: number;
        jCoupling?: Jcoupling;
      }) => number)
  >
>;
export interface GetInternalSignalsOptions {
  /**
   * Initial line width in Hz
   * @default 1
   */
  lineWidth?: number;
  /**
   * frequency
   */
  frequency: number;
  baseline?: number;
  shape?: Shape1D;
  parameters?: ParametersFromOptions;
}

export type Shape1DWithFWHM = Omit<Shape1D, 'fwhm'> & { fwhm: number };
export function getInternalSignals(
  data: DataXY,
  signals: Signal[],
  minMaxY: Record<string, number>,
  options: GetInternalSignalsOptions,
) {
  let index = 0;
  const internalSignals = [];
  const {
    shape: shapeAsOption = { kind: 'gaussian' },
    baseline: shiftValue = minMaxY.min,
    lineWidth = 1,
    frequency,
  } = options;

  const normalizedSignals = normalizeSignals({
    signals,
    data,
    shiftValue,
    frequency,
    range: minMaxY.range,
  });

  for (const signal of normalizedSignals) {
    const { shape = { ...shapeAsOption } } = signal;

    if (!shape.fwhm) {
      shape.fwhm = lineWidth;
    }

    shape.fwhm /= frequency;
    const shapeFct = getShape1D(shape);

    const parameters: Parameter[] = [
      'delta',
      'intensity',
      ...shapeFct.getParameters(),
    ];

    const propertiesValues: Record<string, number[]> = {
      min: [],
      max: [],
      init: [],
      gradientDifference: [],
    };

    for (const parameter of parameters) {
      for (const property of properties) {
        propertiesValues[property].push(
          getPropertyValue({
            signal,
            frequency,
            parameter,
            property,
            minMaxY,
            shapeFct,
          }),
        );
      }
    }

    for (const jCoupling of signal.js || []) {
      for (const property of properties) {
        const parameter = 'coupling' as Parameter;
        const generalParameterValue =
          options?.parameters?.[parameter]?.[property];
        if (generalParameterValue) {
          if (typeof generalParameterValue === 'number') {
            propertiesValues[property].push(
              getNormalizedValue(
                generalParameterValue,
                parameter,
                property,
                minMaxY,
                frequency,
                options.baseline,
              ),
            );
          } else {
            const value = generalParameterValue({ jCoupling });
            propertiesValues[property].push(
              getNormalizedValue(
                value,
                parameter,
                property,
                minMaxY,
                frequency,
                options.baseline,
              ),
            );
          }
        }
        if (!defaultParameters[parameter]) {
          throw new Error(`No default parameter for ${parameter}`);
        }
        const defaultParameterValues = defaultParameters[parameter][property];
        propertiesValues[property].push(defaultParameterValues({ jCoupling }));
      }
    }

    const fromIndex = index;
    const toIndex = fromIndex + parameters.length - 1;

    const fromIndexCoupling = toIndex + 1;
    const toIndexCoupling = fromIndexCoupling + signal.js.length - 1;

    index += toIndexCoupling - fromIndex + 1;
    internalSignals.push({
      shape: shape as Shape1DWithFWHM,
      shapeFct,
      signal,
      parameters,
      propertiesValues,
      fromIndex,
      toIndex,
      fromIndexCoupling,
      toIndexCoupling,
    });
  }

  return internalSignals;
}

function normalizeSignals<T extends Signal>(options: {
  signals: T[];
  data: DataXY;
  shiftValue: number;
  frequency: number;
  range: number;
}) {
  const { signals, data, frequency, shiftValue, range } = options;

  const getMaxIntensity = (signal: Signal) => {
    const peaks = signalsToPointXY([signal], { frequency, maxClusterSize: 1 });
    const biggestPeak = peaks.reduce(
      (biggest, peak) => (peak.y > biggest.y ? peak : biggest),
      peaks[0],
    );
    const { y: maxIntensity } = xyFindClosestPoint(data, biggestPeak.x);
    return maxIntensity;
  };

  return signals.map((signal) => {
    const { intensity = getMaxIntensity(signal) } = signal;
    return {
      ...signal,
      js: signal.js || [],
      intensity: (intensity - shiftValue) / range,
    };
  });
}

function getPropertyValue(options: {
  signal: Signal;
  parameter: Parameter;
  property: string;
  minMaxY: Record<string, number>;
  shapeFct: Shape1DInstance;
  frequency: number;
  baseline?: number;
  parameters?: ParametersFromOptions;
}) {
  const { signal, frequency, parameter, property, minMaxY, shapeFct } = options;

  // check if the property is specified in the signal
  const propertyValue = signal?.parameters?.[parameter]?.[property];
  if (propertyValue) {
    return getNormalizedValue(
      propertyValue,
      parameter,
      property,
      minMaxY,
      frequency,
      options.baseline,
    );
  }

  // check if there are some global option, it could be a number or a callback

  const generalParameterValue = options?.parameters?.[parameter]?.[property];
  if (generalParameterValue) {
    if (typeof generalParameterValue === 'number') {
      return getNormalizedValue(
        generalParameterValue,
        parameter,
        property,
        minMaxY,
        frequency,
        options.baseline,
      );
    } else {
      const value = generalParameterValue(signal);
      return getNormalizedValue(
        value,
        parameter,
        property,
        minMaxY,
        frequency,
        options.baseline,
      );
    }
  }

  // we just need to take the default parameters
  if (!defaultParameters[parameter]) {
    throw new Error(`No default parameter for ${parameter}`);
  }
  const defaultParameterValues = defaultParameters[parameter][property];
  return defaultParameterValues({ signal, shape: shapeFct });
}

function getNormalizedValue(
  value: number,
  parameter: keyof ParametersFromOptions,
  property: string,
  minMaxY: Record<string, number>,
  frequency: number,
  baseline?: number,
) {
  if (parameter === 'intensity') {
    if (property === 'gradientDifference') {
      return value;
    } else {
      return baseline !== undefined
        ? (value - baseline) / minMaxY.range
        : (value - minMaxY.min) / minMaxY.range;
    }
  } else if (parameter === 'fwhm') {
    return value / frequency;
  }
  return value;
}
