import { xSequentialFill } from 'ml-spectra-processing';

import { apodization } from '../apodization/apodization';

import { NMRSignal1D } from './NMRSignal1D';
import { signalsToSpinSystem } from './simulation/signalsToSpinSystem';
import { simulateXYPeaks } from './simulation/simulateXYPeaks';
import { splitSpinSystem } from './simulation/splitSpinSystem';

export interface SignalsToFIDOptions {
  /**
   * at the end of the superposition of all component, an exponential apodization will be applied
   * in order to have FID like shape.
   * @default 1
   */
  lb?: number;
  /**
   * lower limit of the spectrum
   * @default 0
   */
  from?: number;
  /**
   * upper limit of the spectrum
   * @default 10
   */
  to?: number;
  /**
   * number of points of the resulting FID
   */
  nbPoints?: number;
  /**
   * It is the frequency of the carrier (e.g. SFO1 for bruker spectrometers)
   */
  frequency?: number;
  /**
   * Maximum number of atoms on each cluster that can be considered to be simulated together.
   * It affects the the quality and speed of the simulation.
   */
  maxClusterSize?: number;
}

const twoPi = Math.PI * 2;

/**
 * it use the chemical shifts and the coupling between them in order to generate a FID.
 * It does not propagate the hamiltonian of the spin system, instead it simulate all the
 * frequencies in the spectrum (including second order) and add all the waves in a real and
 * imaginary parts, that is why the couplings should be assigned in order to generate the spin system.
 */
export function signalsToFID(
  signals: NMRSignal1D[],
  options: SignalsToFIDOptions,
) {
  const {
    lb = 1,
    from = 0,
    to = 10,
    nbPoints = 1024,
    frequency = 400,
    maxClusterSize = 8,
  } = options;

  const spinSystem = signalsToSpinSystem(signals);

  spinSystem.clusters = splitSpinSystem(spinSystem, {
    frequency,
    maxClusterSize,
  });

  const peaks = simulateXYPeaks(spinSystem, { frequency });
  const re = new Float64Array(nbPoints);
  const im = new Float64Array(nbPoints);

  const sw = Math.abs(to - from);
  const adquisitionTime = nbPoints / (sw * frequency);
  const time = xSequentialFill({
    size: nbPoints,
    from: 0,
    to: adquisitionTime,
  });

  for (const peak of peaks) {
    const { x, y } = peak;
    const cs = x * frequency * twoPi;
    for (let i = 0; i < nbPoints; i++) {
      re[i] += y * Math.cos(cs * time[i]);
      im[i] += y * Math.sin(cs * time[i]);
    }
  }

  const { windowData, ...result } = apodization(
    { re, im },
    {
      pointsToShift: 0,
      compose: {
        length: nbPoints,
        shapes: [
          {
            start: 0,
            shape: {
              kind: 'exponential',
              options: {
                dw: adquisitionTime / (nbPoints - 1),
                lb,
              },
            },
          },
        ],
      },
    },
  );
  return result;
}
