import type { DataXY } from 'cheminfo-types';
import { toBeDeepCloseTo, toMatchCloseTo } from 'jest-matcher-deep-close-to';
import { getShape1D, Shape1D } from 'ml-peak-shape-generator';
import noisyBigPeakSmallPeak from 'nmr-xy-testdata/data/noisy/noisyBigPeakSmallPeak.json';
import tripletQuadruplet from 'nmr-xy-testdata/data/pure/d1-2_j7.json';
import { generateSpectrum } from 'spectrum-generator';

import {
  OptionsXYAutoPeaksPicking,
  xyAutoPeaksPicking,
  XYAutoPeaksPickingWindowSizeError,
} from '../xyAutoPeaksPicking';

expect.extend({ toBeDeepCloseTo, toMatchCloseTo });

describe('xyAutoPeaksPicking', () => {
  it('a triplet and a quadruple', () => {
    const options = { frequency: 1, sgOptions: { windowSize: 9 } };

    const peaks = xyAutoPeaksPicking(tripletQuadruplet, options);

    const widthToFWHM = getShape1D({ kind: 'gaussian' }).widthToFWHM;
    expect(peaks).toHaveLength(7);
    expect(peaks[1].x).toBeDeepCloseTo(0.999831, 3);
    expect(peaks[1].y / 100).toBeDeepCloseTo(14846602893.68, 1);
    expect(peaks[1].width).toBeDeepCloseTo(widthToFWHM(0.0021514892578125), 3);
  });
  it('mixed spectrum with small and big peaks', () => {
    const options = {
      frequency: 1,
      thresholdFactor: 3,
      compile: false,
      clean: false,
      optimize: false,
      minMaxRatio: 0.01,
      integralType: 'sum',
      nH: 3,
      frequencyCluster: 16,
      sgOptions: { windowSize: 9 },
      widthFactor: 4,
      smoothY: false,
      broadWidth: 0.2,
      shape: { kind: 'lorentzian' } as Shape1D,
      broadRatio: 0,
    };

    const peaks = xyAutoPeaksPicking(noisyBigPeakSmallPeak, options);
    const widthToFWHM = getShape1D({ kind: 'lorentzian' }).widthToFWHM;

    const width = 0.4 * options.frequency;
    const expectedResult: Array<Record<string, any>> = [
      { x: 2, y: 6.2683, width, shape: { fwhm: widthToFWHM(width) } },
      { x: 8, y: 318.611, width, shape: { fwhm: widthToFWHM(width) } },
    ];

    expectedResult.forEach((expected, i) => {
      const peak = peaks[i];
      for (const key in expected) {
        //@ts-expect-error the structure is the same
        if (typeof peak[key] === 'object') {
          //@ts-expect-error the structure is the same
          expect(peak[key]).toMatchCloseTo(expected[key], 2);
        } else {
          //@ts-expect-error the structure is the same
          expect(peak[key]).toBeCloseTo(expected[key], 2);
        }
      }
    });
  });

  it('negative spectrum', () => {
    const y = tripletQuadruplet.y;
    for (let i = 0; i < y.length; i++) {
      y[i] *= -1;
    }
    tripletQuadruplet.y = y;
    const options: OptionsXYAutoPeaksPicking = {
      direction: 'negative',
      frequency: 1,
      sgOptions: { windowSize: 9 },
    };
    const peaks = xyAutoPeaksPicking(tripletQuadruplet, options);
    expect(peaks).toHaveLength(7);
    expect(peaks[1].x).toBeDeepCloseTo(0.999831, 3);
    expect(peaks[1].y / 100).toBeDeepCloseTo(-14846602893.68, 1);
    expect(peaks[1].width).toBeDeepCloseTo(0.0021514892578125, 3);
  });

  it('negative and positive data', () => {
    const peaks = [
      { x: -0.5, y: 1, width: 0.05 },
      { x: 0.5, y: -1, width: 0.05 },
    ];

    const data: DataXY = generateSpectrum(peaks, {
      generator: {
        from: -1,
        to: 1,
        nbPoints: 101,
      },
      peakOptions: {
        factor: 6,
      },
    });

    const peakList = xyAutoPeaksPicking(data, {
      frequency: 1,
      direction: 'both',
      smoothY: false,
      sgOptions: { windowSize: 9 },
      noiseLevel: 0,
    });

    const expected: any[] = [
      {
        id: peakList[0].id,
        x: -0.5,
        y: 1,
        width: 0.08,
        shape: {
          kind: 'gaussian',
          fwhm: 0.09419280180123805,
        },
      },
      {
        id: peakList[1].id,
        x: 0.5,
        y: -1,
        width: 0.08,
        shape: {
          kind: 'gaussian',
          fwhm: 0.09419280180123805,
        },
      },
    ];

    expect(peakList).toBeDeepCloseTo(expected);
  });

  it('invalid frequency', () => {
    expect(() =>
      xyAutoPeaksPicking(
        { x: Array.from(Array(20), (v, i) => i), y: [] },
        { frequency: 0 },
      ),
    ).toThrow(XYAutoPeaksPickingWindowSizeError);
  });

  it('invalid data', () => {
    expect(() =>
      xyAutoPeaksPicking(
        { x: [0, undefined as unknown as number], y: [] },
        { frequency: 10 },
      ),
    ).toThrow(XYAutoPeaksPickingWindowSizeError);
  });
});
