import { toBeDeepCloseTo, toMatchCloseTo } from 'jest-matcher-deep-close-to';
import cytisine from 'nmr-xy-testdata/data/experimental/1d/cytisine.json';
import ethylbenzene from 'nmr-xy-testdata/data/experimental/1d/ethylbenzene.json';
import ethylvinylether from 'nmr-xy-testdata/data/experimental/1d/ethylvinylether.json';
import tripletQuadruplet from 'nmr-xy-testdata/data/pure/d1-2_j7.json';
import tenPoints from 'nmr-xy-testdata/data/pure/tenPoints.json';
import twoPoints from 'nmr-xy-testdata/data/pure/twoPoints.json';

import { xyAutoRangesPicking } from '../xyAutoRangesPicking';

expect.extend({ toBeDeepCloseTo, toMatchCloseTo });

const defaultOptions = {
  peakPicking: { frequency: 400, sgOptions: { windowSize: 9 } },
};
describe('xyAutoRangesPicking', () => {
  it('a triplet and a quadruplet', () => {
    const ranges = xyAutoRangesPicking(tripletQuadruplet, defaultOptions);

    expect(ranges).toHaveLength(2);
    expect(ranges[1]).toMatchCloseTo(
      {
        from: 1.96628,
        to: 2.03413,
        integration: 40,
        signals: [
          {
            multiplicity: 'q',
            js: [{ coupling: 7, multiplicity: 'q' }],
            delta: 2,
          },
        ],
      },
      2,
    );
    expect(ranges[0]).toMatchCloseTo(
      {
        from: 0.97473,
        to: 1.02492,
        integration: 60,
        signals: [
          {
            multiplicity: 't',
            js: [{ coupling: 7, multiplicity: 't' }],
            delta: 1,
          },
        ],
      },
      2,
    );
  });
  it('just 2 points', () => {
    const ranges = xyAutoRangesPicking(twoPoints, defaultOptions);
    expect(ranges).toStrictEqual([]);
  });
  it('just 10 points (larger that window', () => {
    const ranges = xyAutoRangesPicking(tenPoints, defaultOptions);
    expect(ranges).toStrictEqual([]);
  });
  it('cytisine', () => {
    const ranges = xyAutoRangesPicking(cytisine, defaultOptions);
    expect(ranges).toHaveLength(11);
  });
  it('ethylvinylether', () => {
    const ranges = xyAutoRangesPicking(ethylvinylether, defaultOptions);
    expect(ranges).toHaveLength(5);
  });
  it('ethylbenzene', () => {
    const ranges = xyAutoRangesPicking(ethylbenzene, defaultOptions);
    expect(ranges).toHaveLength(5);
  });
  it('ethylbenzene first triplet', () => {
    const from = 16500;
    const to = 17200;
    const empty = new Float64Array(100000);
    const ranges = xyAutoRangesPicking(
      {
        x: ethylbenzene.x,
        y: [
          ...empty.slice(0, from),
          ...ethylbenzene.y.slice(from, to),
          ...empty.slice(0, ethylbenzene.x.length - to),
        ],
      },
      defaultOptions,
    );
    expect(ranges).toHaveLength(1);
    expect(ranges[0].signals).toHaveLength(3);
  });
});
