import { PseudoVoigtShape1D } from 'ml-peak-shape-generator';
import { generateSpectrum } from 'spectrum-generator';

import { NMRPeak1DWithShape } from '../../peaks/util/convertWidth';
import { xyPeaksOptimization } from '../xyPeaksOptimization';

describe('optimization of peaks 1D', () => {
  it('two peaks, both direction, gaussians', () => {
    const peaks = [
      { x: -0.5, y: -2, width: 0.12 },
      { x: 0, y: 1, width: 0.12 },
    ];

    const from = -0.7;
    const to = 0.15;
    const frequency = 400;

    const original = generateSpectrum(peaks, {
      generator: {
        from,
        to,
        nbPoints: 256,
        shape: {
          kind: 'gaussian',
        },
      },
    });

    const guessPeaks = [
      {
        x: 0.01,
        y: 0.9,
        width: 0.11 * frequency,
      },
      {
        x: -0.47,
        y: -1.9,
        width: 0.11 * frequency,
      },
    ];

    const optimizedPeaks = xyPeaksOptimization(original, guessPeaks, {
      frequency,
      baseline: 0,
      groupingFactor: 5,
      optimization: {
        kind: 'lm',
        options: {
          maxIterations: 300,
        },
      },
    });

    peaks.sort((a, b) => a.x - b.x);
    for (let i = 0; i < peaks.length; i++) {
      expect(optimizedPeaks[i].x).toBeCloseTo(peaks[i].x);
      expect(optimizedPeaks[i].y).toBeCloseTo(peaks[i].y);
      expect(optimizedPeaks[i].width).toBeCloseTo(peaks[i].width * frequency);
    }
  });
  it('two peaks, both direction, pseudoVoigt', () => {
    const peaks = [
      {
        x: -0.5,
        y: -2,
        width: 0.12,
        shape: { kind: 'pseudoVoigt' as const, mu: 0.1 },
      },
      { x: 0, y: 1, width: 0.12 },
    ];

    const from = -0.7;
    const to = 0.15;
    const frequency = 400;

    const original = generateSpectrum(peaks, {
      generator: {
        from,
        to,
        nbPoints: 256,
      },
    });

    const guessPeaks = [
      {
        x: 0.01,
        y: 0.9,
        width: 0.11 * frequency,
      },
      {
        x: -0.47,
        y: -1.9,
        width: 0.11 * frequency,
      },
    ];

    const optimizedPeaks = xyPeaksOptimization(original, guessPeaks, {
      frequency,
      baseline: 0,
      groupingFactor: 5,
      shape: { kind: 'pseudoVoigt' },
      optimization: {
        kind: 'lm',
        options: {
          maxIterations: 300,
        },
      },
    });

    peaks.sort((a, b) => a.x - b.x);
    for (let i = 0; i < peaks.length; i++) {
      expect(optimizedPeaks[i].x).toBeCloseTo(peaks[i].x);
      expect(optimizedPeaks[i].y).toBeCloseTo(peaks[i].y);
      expect(optimizedPeaks[i].width).toBeCloseTo(peaks[i].width * frequency);
    }
    expect((optimizedPeaks[0].shape as PseudoVoigtShape1D).mu).toBeCloseTo(0.1);
    expect((optimizedPeaks[1].shape as PseudoVoigtShape1D).mu).toBeCloseTo(1);
  });

  it('two peaks, both direction, shape in peaks', () => {
    const peaks = [
      { x: -0.5, y: -2, width: 0.12, shape: { kind: 'lorentzian' as const } },
      { x: 0, y: 1, width: 0.12, shape: { kind: 'gaussian' as const } },
    ];

    const from = -0.7;
    const to = 0.15;
    const frequency = 400;

    const original = generateSpectrum(peaks, {
      generator: {
        from,
        to,
        nbPoints: 256,
      },
    });

    const guessPeaks = [
      {
        x: 0.01,
        y: 0.9,
        width: 0.11 * frequency,
        shape: {
          kind: 'gaussian',
        },
      },
      {
        x: -0.47,
        y: -1.9,
        width: 0.11 * frequency,
        shape: {
          kind: 'lorentzian',
        },
      },
    ];

    const optimizedPeaks = xyPeaksOptimization(original, guessPeaks, {
      frequency,
      baseline: 0,
      groupingFactor: 5,
      shape: { kind: 'pseudoVoigt' }, //should be ignored
      optimization: {
        kind: 'lm',
        options: {
          maxIterations: 300,
        },
      },
    });

    peaks.sort((a, b) => a.x - b.x);
    for (let i = 0; i < peaks.length; i++) {
      expect(optimizedPeaks[i].x).toBeCloseTo(peaks[i].x);
      expect(optimizedPeaks[i].y).toBeCloseTo(peaks[i].y);
      expect(optimizedPeaks[i].width).toBeCloseTo(peaks[i].width * frequency);
    }
    expect((optimizedPeaks[0] as NMRPeak1DWithShape).shape.kind).toBe(
      'lorentzian',
    );
    expect((optimizedPeaks[1] as NMRPeak1DWithShape).shape.kind).toBe(
      'gaussian',
    );
  });
});
